-- Tentative bidi support for babel. Still experimental.
-- (c) 2017 Javier Bezos
-- This version will be activated with a package option named
-- bidi=basic-r and it's simple, for the common cases, and fast.

-- From emacs bidi.c:
-- "Arrrgh!! The UAX#9 algorithm is too deeply entrenched in the
-- assumption of batch-style processing [...]. May the fleas of a
-- thousand camels infest the armpits of those who design supposedly
-- general-purpose algorithms by looking at their own implementations,
-- and fail to consider other possible implementations!" 

-- Well, it took me some time to guess what the batch rules in UAX#9
-- actually mean (in other word, *what* they do and *why*, and not only
-- *how*), but I think (or I hope) I've managed to understand them.

-- From UAX#9: "Where available, markup should be used instead of the
-- explicit formatting characters". So, this simple version just
-- ignores formatting characters. Actually, most of that annex is
-- devoted to how to handle them. 

-- Note setting just the direction in R text is not enough, because
-- there are actually *two* R modes (set explicitly in Unicode with RLM
-- and ALM). In babel the dir is set by a higher protocol based on the
-- language, which in turn sets the correct dir (<l>, <r> or <al>).

-- BD14-BD16 are not implemented. Unicode (and the W3C) are making a
-- great effort to deal with some special problematic cases in
-- "streamed" plain text. I don't think this is the way to go --
-- particular issues should be fixed by a high level interface taking
-- into account the needs of the document. And here is where luatex
-- excels, because everything related to bidi writing is under our
-- control.

----------------------------

-- Must be set by babel:

langdirs = {}

log = texio.write_nl
function log(s) end

require('bidi-chars.lua')

local DIR = node.id("dir")

function dir_mark(head, from, to, outer)
  dir = (outer == 'r') and 'TLT' or 'TRT' -- ie, reverse
  log('====') ----
  local d = node.new(DIR)
  d.dir = '+' .. dir
  node.insert_before(head, from, d)
  d = node.new(DIR)
  d.dir = '-' .. dir
  node.insert_after(head, to, d)
end

function bidi_text (head)
  local first_n, last_n = nil, nil -- first and last char with nums
  local last_es = nil              -- an auxiliary 'last' used with nums
  local first_d, last_d = nil, nil -- first and last char in L/R block
  local dir = nil
  local dir_real = nil

  -- Next also depends on script/lang (<al>/<r>). To be set by babel.
  -- tex.pardir is dangerous, could be (re)set -- but it should be
  -- changed only in vmode. There are two strong's -- strong = l/al/r
  -- and strong_lr = l/r (there must be a better way):

  local strong = ('TRT' == tex.pardir) and 'r' or 'l'  
  local strong_lr = (strong == 'l') and 'l' or 'r' 
  local outer = strong

  local new_dir = false
  local first_dir = false

  local last_lr = nil

  local an = false -----

  log('***' .. strong .. strong_lr)

  for item in node.traverse(head) do

    if item.id == node.id'glyph' then
      chardata = characters.data[item.char]
      dir = chardata and chardata.direction or 'l'

      -- Next is based on the assumption babel sets the language AND
      -- switches the script with its dir. We treat a language block
      -- as a separate Unicode sequence. The following piece of code
      -- is executed at the first glyph after a 'dir' node. We don't
      -- know the current language until then.

      if new_dir then
        strong = langdirs[item.lang]
        strong_lr = (strong == 'l') and 'l' or 'r' 
        outer = strong_lr
        new_dir = false
      end

      -- The dual <al>/<r> system for R is somewhat cumbersome.

      dir_real = dir               -- We need dir_real to set strong below
      if dir == 'al' then dir = 'r' end -- W3

      -- W1:

      if dir == 'nsm' then dir = strong end

      -- By W2, there are no <en> <et> <es> if strong == <al>, only
      -- <an>. Therefore, there are not <et en> nor <en et>, W5
      -- can be ignored, and W6 applied:

      if strong == 'al' then
        if dir == 'en' then dir = 'an' end                -- W2
        if dir == 'et' or dir == 'es' then dir = 'on' end -- W6
        strong_lr = 'r'                                   -- W3
      end
      log(item.char .. ' == ' .. dir .. ' << ' .. strong .. ' ** ' .. outer)
    elseif item.id == node.id'dir' then
      new_dir = true 
      dir = nil
      log('>< ' .. strong .. (item.dir or '')) ---------
    else
      dir = nil          -- Not a char
      log('?? ' .. strong) ---------
    end

    -- Numbers in R mode. A sequence of <en>, <et>, <an>, <es> and
    -- <cs> is typeset (with some rules) in L mode. We store the
    -- starting and ending points, and only when anything different is
    -- found (including nil, ie, a non-char), the textdir is set. This
    -- means you cannot insert, say, a whatsit, but this is what I
    -- would expect (with luacolor you may colorize some
    -- digits). Anyway, this behaviour could be changed with a switch
    -- in the future.  Note in the first branch only <an> is relevant
    -- if <al>.

      if dir == 'an' and first_d and strong_lr == 'l' then
      	    -- TODO - Very ugly. I'm not sure it's the right way
	    -- <an> is after all Arabic, so an L mode stops
	    -- but what about <en> in Hebrew?
        dir_mark(head, first_d, last_d, outer)
        first_d, last_d = nil, nil
	an = true
      end

    if strong_lr == 'r' or an then
      if dir == 'en' or dir == 'et' or dir == 'an' then
        first_n = first_n or item
        last_n = last_es or item
        last_es = nil
      elseif dir == 'es' and last_n then -- W3+W6
        last_es = item
      elseif dir == 'cs' then            -- it's right - do nothing
      elseif first_n then -- & if dir = any but en, et, an, es, cs, inc nil
        dir_mark(head, first_n, last_n, 'r')
        first_n, last_n = nil, nil
	an = false
      end
    end
   
    -- R text in L, or L text in R. Order of dir_ mark's are relevant:
    -- d goes outside n, and therefore it's emitted after. See
    -- dir_mark to understand why (but is the nesting actually
    -- necessary or is a flat dir structure enough?). Only L, R (and
    -- AL) chars are taken into account -- anything else, including
    -- spaces, whatsits, etc., are ignored:

    if (dir == 'l' or dir == 'r') then
      if dir ~= outer then
        first_d = first_d or item
        last_d = item
      elseif first_d and dir ~= strong_lr then
        dir_mark(head, first_d, last_d, outer)
        first_d, last_d = nil, nil
     end
    end
    
    -- Mirroring

    -- Each chunk of text in a certain language is considered a
    -- "closed" sequence.  If <r on r> and <l on l>, it's clearly <r>
    -- and <l>, resptly, but with other combinations depends on
    -- outer. From all these, we select only those resolving <on> ->
    -- <r>. At the beginning (when last_lr is nil) of an R text, they
    -- are mirrored directly.

    -- TODO - numbers in R mode are processed. It doesn't hurt, but
    -- should not be done.

    if dir and not last_lr and dir ~= 'l' and outer == 'r' then
      log('@( ' .. item.char) -------
      item.char = characters.data[item.char].mirror or item.char
    elseif (dir or new_dir) and last_lr ~= item then
      local mir = outer .. strong_lr .. (dir or outer)
      if mir == 'rrr' or mir == 'lrr' or mir == 'rrl' or mir == 'rlr' then
	 --string.match(outer .. strong_lr .. (dir or outer), '????')
         -- ie, if rrr, lrr, rrl, rlr, lrr, but not lll, rll, lrl, llr
        for ch in node.traverse(node.next(last_lr)) do
          if ch == item then break end
	  if ch.id == node.id'glyph' then
	    log('@ ' .. ch.char) -------
	    ch.char = characters.data[ch.char].mirror or ch.char
	    log('@> ' .. ch.char) ------
	  end
        end
        --last_lr = nil
      end
    end

    -- Save some values for the next iteration. If the current node is
    -- 'dir', open a new sequence. Since dir could be changed, strong
    -- is set with its real value (dir_real). 

    if dir == 'l' or dir == 'r' then
      last_lr = item
      strong = dir_real            -- Don't search back - best save now
      strong_lr = (strong == 'l') and 'l' or 'r'
    elseif new_dir then
      last_lr = nil
    end

  end 

  -- Mirror the last chars if they are no directed. And make sure any
  -- open block is closed, too.

  if last_lr and outer == 'r' then
    for ch in node.traverse_id(node.id'glyph', node.next(last_lr)) do
      log('@) ' .. ch.char) -------
      ch.char = characters.data[ch.char].mirror or ch.char
    end
  end
  if first_n then 
    dir_mark(head, first_n, last_n, outer)
  end
  if first_d then 
    dir_mark(head, first_d, last_d, outer)
  end

  -- In boxes, the dir node could be added before the original head, so
  -- the actual head is the previous node.
  
  return node.prev(head) or head
end
